#  Copyright (c) 2023 zfit

import mplhep
import numpy as np
import numpy.testing
import pytest
from matplotlib import pyplot as plt

import zfit
import zfit.z.numpy as znp
from zfit._data.binneddatav1 import BinnedData
from zfit._loss.binnedloss import ExtendedBinnedNLL
from zfit.models.binned_functor import BinnedSumPDF
from zfit.models.morphing import SplineMorphingPDF
from zfit.models.template import BinnedTemplatePDFV1


def test_binned_template_pdf():
    bins1 = 5
    bins2 = 7
    counts = np.random.uniform(high=1, size=(bins1, bins2))  # generate counts
    counts2 = np.random.normal(loc=5, size=(bins1, bins2))
    counts3 = (
        np.linspace(0, 10, num=bins1)[:, None] * np.linspace(0, 5, num=bins2)[None, :]
    )
    binnings = [
        zfit.binned.RegularBinning(bins1, 0, 10, name="obs1"),
        zfit.binned.RegularBinning(7, -10, bins2, name="obs2"),
    ]
    binning = binnings
    axes = zfit.binned.Binnings(binning)
    obs = zfit.Space(obs=["obs1", "obs2"], binning=binning)

    data = BinnedData.from_tensor(
        space=obs, values=counts, variances=znp.ones_like(counts) * 1.3
    )
    data2 = BinnedData.from_tensor(obs, counts2)
    data3 = BinnedData.from_tensor(obs, counts3)

    pdf = BinnedTemplatePDFV1(data=data, extended=np.sum(counts))
    pdf2 = BinnedTemplatePDFV1(data=data2, extended=np.sum(counts2))
    pdf3 = BinnedTemplatePDFV1(data=data3, extended=np.sum(counts3))
    assert len(pdf.ext_pdf(data)) > 0
    pdf_sum = BinnedSumPDF(pdfs=[pdf, pdf2, pdf3], obs=obs)

    probs = pdf_sum.counts(data)
    true_sum_counts = counts + counts2 + counts3
    np.testing.assert_allclose(true_sum_counts, probs)
    nsamples = 100_000_000
    sample = pdf_sum.sample(n=nsamples)
    np.testing.assert_allclose(
        true_sum_counts, sample.values() / nsamples * pdf_sum.get_yield(), rtol=0.03
    )

    # integrate
    true_integral = znp.sum(true_sum_counts)
    integral = pdf_sum.ext_integrate(limits=obs)
    assert pytest.approx(float(true_integral)) == float(integral)

    # import matplotlib.pyplot as plt
    # plt.imshow(probs)
    # plt.imshow(counts2)
    # plt.show()
    # assert len(pdf.pdf(None, obs)) > 0


@pytest.mark.plots
@pytest.mark.parametrize("alphas", [None, [-0.7, -0.1, 0.5, 1.4]])
def test_morphing_templates(alphas):
    bins1 = 15
    irregular_str = "irregular templates" if alphas is not None else ""

    counts1 = np.random.uniform(70, high=100, size=bins1)  # generate counts
    counts = [
        counts1 - np.random.uniform(high=20, size=bins1),
        counts1,
        counts1 + np.random.uniform(high=20, size=bins1),
    ]
    if alphas is not None:
        counts.append(counts1 + np.random.uniform(high=5, size=bins1))
    binning = zfit.binned.RegularBinning(bins1, 0, 10, name="obs1")
    obs = zfit.Space(obs="obs1", binning=binning)
    histos = [BinnedData.from_tensor(obs, count) for count in counts]
    pdfs = [zfit.pdf.HistogramPDF(h) for h in histos]
    if alphas is not None:
        pdfs = {a: p for a, p in zip(alphas, pdfs)}
    alpha = zfit.Parameter("alpha", 0, -5, 5)
    morph = SplineMorphingPDF(alpha=alpha, hists=pdfs)
    if alphas is None:
        alphas = [-1, 0, 1]
    for i, a in enumerate(alphas):
        alpha.set_value(a)
        np.testing.assert_allclose(morph.counts(), counts[i])
        if len(alphas) > i + 1:
            alpha.set_value((a + alphas[i + 1]) / 2)
            max_dist = (counts[i] - counts[i + 1]) ** 2 + 5  # tolerance
            max_dist *= 1.1  # not strict, it can be a bit higher
            numpy.testing.assert_array_less((morph.counts() - counts[i]) ** 2, max_dist)
            numpy.testing.assert_array_less(
                (morph.counts() - counts[i + 1]) ** 2, max_dist
            )

    import matplotlib.cm as cm

    amin, amax = -2, 2
    n = 5

    template_alphas = np.array(list(alphas))

    for do_3d in [True, False]:
        plt.figure()
        if do_3d:
            ax = plt.gcf().add_subplot(111, projection="3d")
        else:
            ax = plt.gca()
        plotstyle = "3d plot" if do_3d else "hist plot"
        plt.title(f"Morphing with splines {irregular_str} {plotstyle}")

        for a in list(znp.linspace(amin, amax, n * 2)) + list(template_alphas):
            normed_a = (a - amin) / (amax - amin) / 1.3  # 3 is a scaling factor
            color = cm.get_cmap("winter")(normed_a)
            alpha.set_value(a)
            histo = morph.ext_pdf(None)
            histo = BinnedData.from_tensor(obs, histo)
            histo = histo.to_hist()
            values = histo.values()
            x = histo.axes.edges[0][:-1]
            y = np.broadcast_to(a, values.shape)
            z = values
            label = None
            if do_3d:
                ax.step(x, y, z, color=color, where="pre", label=label)
            else:
                if np.min((a - template_alphas) ** 2) < 0.0001:
                    label = f"alpha={a}"
                mplhep.histplot(histo, label=label, color=color)
        ax.set_xlabel("observable")
        ax.set_ylabel("alpha")
        if do_3d:
            ax.set_zlabel("ext_pdf")
        plt.legend()
        pytest.zfit_savefig()


@pytest.mark.parametrize("alphas", [None, [-0.7, -0.1, 0.5, 1.4]])
def test_morphing_templates2D(alphas):
    bins1 = 10
    bins2 = 7
    shape = (bins1, bins2)
    counts1 = np.random.uniform(70, high=100, size=shape)  # generate counts
    # counts1 = np.random.uniform(70, high=100, size=bins1)  # generate counts
    counts = [
        counts1 - np.random.uniform(high=20, size=shape),
        counts1,
        counts1 + np.random.uniform(high=20, size=shape),
    ]
    if alphas is not None:
        counts.append(counts1 + np.random.uniform(high=5, size=shape))
    binning1 = zfit.binned.VariableBinning(
        sorted(np.random.uniform(0, 10, size=bins1 + 1)), name="obs1"
    )
    binning2 = zfit.binned.RegularBinning(bins2, 0, 10, name="obs2")
    obs1 = zfit.Space(obs="obs1", binning=binning1)
    obs2 = zfit.Space(obs="obs2", binning=binning2)
    obs = obs1 * obs2
    datasets = [BinnedData.from_tensor(obs, count) for count in counts]
    pdfs = [
        BinnedTemplatePDFV1(data=data, extended=np.sum(data.values()))
        for data in datasets
    ]
    if alphas is not None:
        pdfs = {a: p for a, p in zip(alphas, pdfs)}
    alpha = zfit.Parameter("alpha", 0, -5, 5)
    morph = SplineMorphingPDF(alpha=alpha, hists=pdfs)
    if alphas is None:
        alphas = [-1, 0, 1]
    for i, a in enumerate(alphas):
        alpha.set_value(a)
        np.testing.assert_allclose(morph.counts(), counts[i])
        assert pytest.approx(np.sum(counts[i])) == zfit.run(morph.get_yield())
        if len(alphas) > i + 1:
            alpha.set_value((a + alphas[i + 1]) / 2)
            max_dist = (counts[i] - counts[i + 1]) ** 2 + 5  # tolerance
            max_dist *= 1.1  # not strict, it can be a bit higher
            numpy.testing.assert_array_less((morph.counts() - counts[i]) ** 2, max_dist)
            numpy.testing.assert_array_less(
                (morph.counts() - counts[i + 1]) ** 2, max_dist
            )


def _binned_template_composed_factory(data, sysshape):
    pdf = zfit.pdf.HistogramPDF(data=data, extended=True)
    pdf = zfit.pdf.BinwiseScaleModifier(pdf=pdf, modifiers=sysshape)
    return pdf


@pytest.mark.parametrize(
    "TemplateLikePDF", [BinnedTemplatePDFV1, _binned_template_composed_factory]
)
def test_binned_template_pdf_bbfull(TemplateLikePDF):
    bins1 = 7
    bins2 = 5

    counts1 = np.random.uniform(high=150, size=(bins1, bins2))  # generate counts
    counts2 = np.random.normal(loc=50, size=(bins1, bins2))
    counts3 = (
        np.linspace(10, 100, num=bins1)[:, None]
        * np.linspace(10, 500, num=bins2)[None, :]
    )
    binnings = [
        zfit.binned.RegularBinning(bins1, 0, 10, name="obs1"),
        zfit.binned.RegularBinning(bins2, -10, 7, name="obs2"),
    ]
    binning = binnings
    obs = zfit.Space(obs=["obs1", "obs2"], binning=binning)

    mc1 = BinnedData.from_tensor(
        space=obs, values=counts1, variances=znp.ones_like(counts1) * 1.3
    )
    mc2 = BinnedData.from_tensor(obs, counts2)
    mc3 = BinnedData.from_tensor(obs, counts3)

    counts_mc = counts1 + counts2 + counts3

    counts1_data = np.random.uniform(high=150, size=(bins1, bins2))  # generate counts
    counts2_data = np.random.normal(loc=50, size=(bins1, bins2))
    counts3_data = (
        np.linspace(10, 100, num=bins1)[:, None]
        * np.linspace(20, 490, num=bins2)[None, :]
    )
    counts_data = counts1_data + counts2_data + counts3_data
    counts_data *= 1.1
    data = BinnedData.from_tensor(space=obs, values=counts_data)

    pdf1 = TemplateLikePDF(data=mc1, sysshape=True)
    pdf2 = TemplateLikePDF(data=mc2, sysshape=True)
    pdf3 = TemplateLikePDF(data=mc3, sysshape=True)
    assert len(pdf1.counts()) > 0
    pdf_sum = BinnedSumPDF(pdfs=[pdf1, pdf2, pdf3], obs=obs)
    counts1_flat = np.reshape(counts1, -1)
    constraints1 = zfit.constraint.GaussianConstraint(
        pdf1.params.values(),
        observation=np.ones_like(counts1_flat),
        uncertainty=np.sqrt(counts1_flat) / counts1_flat,
    )
    counts2_flat = np.reshape(counts2, -1)
    constraints2 = zfit.constraint.GaussianConstraint(
        pdf2.params.values(),
        observation=np.ones_like(counts2_flat),
        uncertainty=np.sqrt(counts2_flat) / counts2_flat,
    )
    counts3_flat = np.reshape(counts3, -1)
    constraints3 = zfit.constraint.GaussianConstraint(
        pdf3.params.values(),
        observation=np.ones_like(counts3_flat),
        uncertainty=np.sqrt(counts3_flat) / counts3_flat,
    )
    # constraints2 = zfit.constraint.PoissonConstraint(pdf2.params.values(), np.reshape(counts2, -1))
    # constraints3 = zfit.constraint.PoissonConstraint(pdf3.params.values(), np.reshape(counts3, -1))
    loss = ExtendedBinnedNLL(
        pdf_sum, data, constraints=[constraints1, constraints2, constraints3]
    )
    # for i in progressbar.progressbar(range(1000000)):
    loss.value(full=False)
    loss.gradient()
    minimizer = zfit.minimize.Minuit(gradient=False, mode=0)
    # minimizer = zfit.minimize.ScipyTrustConstrV1(verbosity=8)
    minimizer.minimize(loss)

    counts = pdf_sum.counts()
    np.testing.assert_array_less(
        counts_mc, counts_data
    )  # this is an assumption, if it's is wrong, the test is flawed
    np.testing.assert_array_less(counts, counts_data)
    np.testing.assert_array_less(counts_mc, counts)
    # np.testing.assert_allclose(counts_data, probs, rtol=0.01)



"""A rich selection of analytically implemented Distributions (models) are available in `TensorFlow Probability.

<https://github.com/tensorflow/probability>`_. While their API is slightly different from the zfit models, it is similar
enough to be easily wrapped.

Therefore, a convenient wrapper as well as a lot of implementations are provided.
"""

#  Copyright (c) 2024 zfit
from __future__ import annotations

from collections import OrderedDict

import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_probability.python.distributions as tfd
from pydantic import Field

from typing import Literal

from zfit import z
from zfit.util.exception import (
    AnalyticSamplingNotImplemented,
)
from ..core.basepdf import BasePDF
from ..core.interfaces import ZfitData
from ..core.parameter import convert_to_parameter
from ..core.serialmixin import SerializableMixin
from ..core.space import Space, supports
from ..serialization import SpaceRepr, Serializer
from ..serialization.pdfrepr import BasePDFRepr
from ..settings import ztypes
from ..util import ztyping
from ..util.deprecation import deprecated_args
from ..util.ztyping import ExtendedInputType, NormInputType


# TODO: improve? while loop over `.sample`? Maybe as a fallback if not implemented?


def tfd_analytic_sample(n: int, dist: tfd.Distribution, limits: ztyping.ObsTypeInput):
    """Sample analytically with a `tfd.Distribution` within the limits. No preprocessing.

    Args:
        n: Number of samples to get
        dist: Distribution to sample from
        limits: Limits to sample from within

    Returns:
        The sampled data with the number of samples and the number of observables.
    """
    lower_bound, upper_bound = limits.rect_limits
    lower_prob_lim = dist.cdf(lower_bound)
    upper_prob_lim = dist.cdf(upper_bound)

    shape = (n, 1)
    prob_sample = z.random.uniform(
        shape=shape, minval=lower_prob_lim, maxval=upper_prob_lim
    )
    prob_sample.set_shape((None, 1))
    try:
        sample = dist.quantile(prob_sample)
    except NotImplementedError:
        raise AnalyticSamplingNotImplemented
    sample.set_shape((None, limits.n_obs))
    return sample


class WrapDistribution(BasePDF):  # TODO: extend functionality of wrapper, like icdf
    """Baseclass to wrap tensorflow-probability distributions automatically."""

    def __init__(
        self,
        distribution,
        dist_params,
        obs,
        params=None,
        dist_kwargs=None,
        dtype=ztypes.float,
        name=None,
        **kwargs,
    ):
        # Check if subclass of distribution?
        if dist_kwargs is None:
            dist_kwargs = {}

        if dist_params is None:
            dist_params = {}
        name = name or distribution.name
        if params is None:
            params = OrderedDict((k, p) for k, p in dist_params.items())
        else:
            params = OrderedDict(
                (k, convert_to_parameter(p)) for k, p in params.items()
            )

        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)

        self._distribution = distribution
        self.dist_params = dist_params
        self.dist_kwargs = dist_kwargs
        self._inverse_analytic_integral = []

    @property
    def distribution(self):
        params = self.dist_params
        if callable(params):
            params = params()
        kwargs = self.dist_kwargs
        if callable(kwargs):
            kwargs = kwargs()
        return self._distribution(**params, **kwargs, name=self.name + "_tfp")

    def _unnormalized_pdf(self, x: ZfitData):
        value = z.unstack_x(x)  # TODO: use this? change shaping below?
        return self.distribution.prob(value=value, name="unnormalized_pdf")

    # TODO: register integral?
    @supports()
    def _analytic_integrate(self, limits, norm):
        lower, upper = limits._rect_limits_tf
        lower = z.unstack_x(lower)
        upper = z.unstack_x(upper)
        tf.debugging.assert_all_finite(
            (lower, upper), "Are infinite limits needed? Causes troubles with NaNs"
        )
        return self.distribution.cdf(upper) - self.distribution.cdf(lower)

    def _analytic_sample(self, n, limits: Space):
        return tfd_analytic_sample(n=n, dist=self.distribution, limits=limits)


# class KernelDensityTFP(WrapDistribution):
#
#     def __init__(self, loc: ztyping.ParamTypeInput, scale: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput,
#                  kernel: tfp.distributions.Distribution = tfp.distributions.Normal,
#                  weights: Union[None, np.ndarray, tf.Tensor] = None, name: str = "KernelDensity"):
#         """Kernel Density Estimation of loc and either a broadcasted or a per-loc scale with a Distribution as kernel.
#
#         Args:
#             loc: 1-D Tensor-like. The positions of the `kernel`. Determines how many kernels will be created.
#             scale: Broadcastable to the batch and event shape of the distribution. A scalar will simply broadcast
#                 to `loc` for a 1-D distribution.
#             obs: Observables
#             kernel: Distribution that is used as kernel
#             weights: Weights of each `loc`, can be None or Tensor-like with shape compatible with loc
#             name: Name of the PDF
#         """
#         if not isinstance(kernel,
#                           tfp.distributions.Distribution) and False:  # HACK remove False, why does test not work?
#             raise TypeError("Currently, only tfp distributions are supported as kernels. Please open an issue if this "
#                             "is too restrictive.")
#
#         if isinstance(loc, ZfitData):
#             if loc.weights is not None:
#                 if weights is not None:
#                     raise OverdefinedError("Cannot specify weights and use a `ZfitData` with weights.")
#                 else:
#                     weights = loc.weights
#
#         if weights is None:
#             weights = tf.ones_like(loc, dtype=tf.float64)
#         self._weights_loc = weights
#         self._weights_sum = z.reduce_sum(weights)
#         self._latent_loc = loc
#         params = {"scale": scale}
#         dist_params = {"loc": loc, "scale": scale}
#         super().__init__(distribution=kernel, dist_params=dist_params, obs=obs, params=params, dtype=ztypes.float,
#                          name=name)
#
#     def _unnormalized_pdf(self, x: "zfit.Data", norm_range=False):
#         value = znp.expand_dims(x.value(), -2)
#         new_shape = znp.concatenate([tf.shape(value)[:2], [tf.shape(self._latent_loc)[0], 4]], axis=0)
#         value = tf.broadcast_to(value, new_shape)
#         probs = self.distribution.prob(value=value, name="unnormalized_pdf")
#         # weights = znp.expand_dims(self._weights_loc, axis=-1)
#         weights = self._weights_loc
#         probs = z.reduce_sum(probs * weights, axis=-1) / self._weights_sum
#         return probs
#
#     @supports()
#     def _analytic_integrate(self, limits, norm_range):
#         lower, upper = limits.limits
#         if np.all(-np.array(lower) == np.array(upper)) and np.all(np.array(upper) == np.infty):
#             return z.reduce_sum(self._weights_loc)  # tfp distributions are normalized to 1
#         lower = z.to_real(lower[0], dtype=self.dtype)
#         # lower = tf.broadcast_to(lower, shape=(tf.shape(self._latent_loc)[0], limits.n_obs,))  # remove
#         upper = z.to_real(upper[0], dtype=self.dtype)
#         integral = self.distribution.cdf(upper) - self.distribution.cdf(lower)
#         integral = z.reduce_sum(integral * self._weights_loc, axis=-1) / self._weights_sum
#         return integral  # TODO: generalize for VectorSpaces


class Gauss(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Gauss",
    ):
        """Gaussian or Normal distribution with a mean (mu) and a standartdeviation (sigma).

        The gaussian shape is defined as

        .. math::
            f(x \\mid \\mu, \\sigma^2) = e^{ -\\frac{(x - \\mu)^{2}}{2\\sigma^2} }

        with the normalization over [-inf, inf] of

        .. math::
            \\frac{1}{\\sqrt{2\\pi\\sigma^2} }

        The normalization changes for different normalization ranges

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
        """
        mu, sigma = self._check_input_params(mu, sigma)
        params = OrderedDict((("mu", mu), ("sigma", sigma)))
        dist_params = lambda: dict(loc=mu.value(), scale=sigma.value())
        distribution = tfp.distributions.Normal
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
        )


class GaussPDFRepr(BasePDFRepr):
    _implementation = Gauss
    hs3_type: Literal["Gauss"] = Field("Gauss", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamInputTypeDiscriminated
    sigma: Serializer.types.ParamInputTypeDiscriminated


class ExponentialTFP(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        tau: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        name: str = "Exponential",
    ):
        (tau,) = self._check_input_params(tau)
        params = OrderedDict((("tau", tau),))
        dist_params = dict(rate=tau)
        distribution = tfp.distributions.Exponential
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
        )


class Uniform(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        low: ztyping.ParamTypeInput,
        high: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Uniform",
    ):
        """Uniform distribution which is constant between `low`, `high` and zero outside.

        Args:
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.pdf||@docend:model.init.pdf|
        """
        low, high = self._check_input_params(low, high)
        params = OrderedDict((("low", low), ("high", high)))
        dist_params = lambda: dict(low=low.value(), high=high.value())
        distribution = tfp.distributions.Uniform
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
        )


class TruncatedGauss(WrapDistribution):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        low: ztyping.ParamTypeInput,
        high: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "TruncatedGauss",
    ):
        """Gaussian distribution that is 0 outside of `low`, `high`. Equivalent to the product of Gauss and Uniform.

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
        """
        mu, sigma, low, high = self._check_input_params(mu, sigma, low, high)
        params = OrderedDict(
            (("mu", mu), ("sigma", sigma), ("low", low), ("high", high))
        )
        distribution = tfp.distributions.TruncatedNormal
        dist_params = lambda: dict(
            loc=mu.value(), scale=sigma.value(), low=low.value(), high=high.value()
        )
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
        )


class Cauchy(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        m: ztyping.ParamTypeInput,
        gamma: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Cauchy",
    ):
        r"""Non-relativistic Breit-Wigner (Cauchy) PDF representing the energy distribution of a decaying particle.

        The (unnormalized) shape of the non-relativistic Breit-Wigner is given by

        .. math::

            \frac{1}{\gamma \left[1 + \left(\frac{x - m}{\gamma}\right)^2\right]}

        with :math:`m` the mean and :math:`\gamma` the width of the distribution.

        Args:
            m: Invariant mass of the unstable particle.
            gamma: Width of the shape.
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:model.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:model.init.name|
        """
        m, gamma = self._check_input_params(m, gamma)
        params = OrderedDict((("m", m), ("gamma", gamma)))
        distribution = tfp.distributions.Cauchy
        dist_params = lambda: dict(loc=m.value(), scale=gamma.value())
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
        )


class CauchyPDFRepr(BasePDFRepr):
    _implementation = Cauchy
    hs3_type: Literal["Cauchy"] = Field("Cauchy", alias="type")
    x: SpaceRepr
    m: Serializer.types.ParamTypeDiscriminated
    gamma: Serializer.types.ParamTypeDiscriminated


class Poisson(WrapDistribution, SerializableMixin):
    _N_OBS = 1

    @deprecated_args(None, "Use lam instead", "lamb")
    def __init__(
        self,
        lam: ztyping.ParamTypeInput = None,
        obs: ztyping.ObsTypeInput = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Poisson",
        lamb=None,
    ):
        """Poisson distribution, parametrized with an event rate parameter (lamb).

        The probability mass function of the Poisson distribution is given by

        .. math::
            f(x, \\lambda) = \\frac{\\lambda^{x}e^{-\\lambda}}{x!}

        Args:
            lamb: the event rate
            obs: Observables and normalization range the pdf is defined in
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: Name of the PDF
        """
        if lamb is not None:
            lam = lamb
        del lamb
        (lam,) = self._check_input_params(lam)
        params = {"lam": lam}
        dist_params = lambda: dict(rate=lam.value())
        distribution = tfp.distributions.Poisson
        super().__init__(
            distribution=distribution,
            dist_params=dist_params,
            obs=obs,
            params=params,
            name=name,
            extended=extended,
            norm=norm,
        )


class PoissonPDFRepr(BasePDFRepr):
    _implementation = Poisson
    hs3_type: Literal["Poisson"] = Field("Poisson", alias="type")
    x: SpaceRepr
    lam: Serializer.types.ParamTypeDiscriminated



#  Copyright (c) 2023 zfit
import numpy as np
import pytest
import scipy.stats

import zfit
from zfit import z
from zfit.z.math import (
    autodiff_gradient,
    autodiff_hessian,
    numerical_gradient,
    numerical_hessian,
)


def test_numerical_gradient():
    param1 = zfit.Parameter("param1", 4.0)
    param2 = zfit.Parameter("param2", 5.0)
    param3 = zfit.Parameter("param3", 2.0)

    def func1():
        return param1 * param2**2 + param3**param1

    num_gradients = numerical_gradient(func1, params=[param1, param2, param3])
    tf_gradients = autodiff_gradient(func1, params=[param1, param2, param3])
    np.testing.assert_allclose(num_gradients, tf_gradients)


@pytest.mark.parametrize("graph", [False, True])
def test_numerical_hessian(graph):
    param1 = zfit.Parameter("param1", 4.0)
    param2 = zfit.Parameter("param2", 5.0)
    param3 = zfit.Parameter("param3", 2.0)

    def func1():
        return param1 * param2**2 + param3**param1

    def create_derivatives(func1, params):
        num_hessian_diag = numerical_hessian(func1, params=params, hessian="diag")
        num_hessian = numerical_hessian(func1, params=params)
        tf_hessian_diag = autodiff_hessian(func1, params=params, hessian="diag")
        tf_hessian = autodiff_hessian(func1, params=params)
        return num_hessian, num_hessian_diag, tf_hessian, tf_hessian_diag

    params = [param1, param2, param3]
    if graph:
        create_derivatives = z.function(create_derivatives)
    num_hessian, num_hessian_diag, tf_hessian, tf_hessian_diag = create_derivatives(
        func1, params
    )
    np.testing.assert_allclose(num_hessian, tf_hessian, rtol=3e-4, atol=1e-8)
    tf_hessian_diag_from_hessian = [tf_hessian[i, i] for i in range(len(params))]
    np.testing.assert_allclose(
        tf_hessian_diag_from_hessian, tf_hessian_diag, rtol=3e-4, atol=1e-8
    )
    np.testing.assert_allclose(num_hessian_diag, tf_hessian_diag, rtol=3e-4, atol=1e-8)


def test_reduce_geometric_mean():
    import zfit.z.numpy as znp

    rnd1 = np.random.poisson(1000, size=(54, 14, 3)).astype(np.float64)
    gmean_np = scipy.stats.mstats.gmean(rnd1, axis=None)
    gmea_z = z.math.reduce_geometric_mean(rnd1)
    gmea_z_weighted = z.math.reduce_geometric_mean(rnd1, weights=znp.ones_like(rnd1))
    np.testing.assert_allclose(gmea_z, gmean_np)
    np.testing.assert_allclose(gmea_z_weighted, gmean_np)



"""Module for caching.

The basic concept of caching in Zfit builds on a "cacher", that caches a certain value and that
is dependent of "cache_dependents". By implementing `ZfitGraphCachable`, an object will be able to play both
roles. And most importantly, it has a `_cache` dict, that contains all the cache.

Basic principle
===============

A "cacher" adds any dependents that it may comes across with `add_cache_dependents`. For example,
for a loss this would be all pdfs and data. Since :py:class:`~zfit.Space` is immutable, there is no need to add this
as a dependent. This leads to the "cache_dependent" to register the "cacher" and to remember it.

In case, any "cache_dependent" changes in a way the cache of itself (and any "cacher") is invalid,
which is done in the simplest case by decorating a method with `@invalidates_cache`, the "cache_dependent":

 * clears it's own cache with `reset_cache_self` and
 * "clears" any "cacher"s cache with `reset_cache(reseter=self)`, telling the "cacher" that it should
   reset the cache. This is also where more fine-grained control (depending on which "cache_dependent"
   calls `reset_cache`) can be brought into play.

Example with a pdf that caches the normalization:

.. code:: python

    class Parameter(Cachable):
        def load(new_value):  # does not require to build a new graph
            # do something

        @invalidates_cache
        def change_limits(new_limits):  # requires to build a new graph (as an example)
            # do something

    # create param1, param2 from `Parameter`

    class MyPDF(Cachable):
        def __init__(self, param1, param2):
            self.add_cache_dependents([param1, param2])

        def cached_func(...):
            if self._cache.get('my_name') is None:
                result = ...  # calculations here
                self._cache['my_name']
            else:
                result = self._cache['my_name']
            return result
"""

#  Copyright (c) 2024 zfit

from __future__ import annotations

import collections.abc
import functools
import weakref
from abc import abstractmethod
from collections.abc import Iterable
from itertools import zip_longest

import numpy as np
import tensorflow as tf
import uhi.typing.plottable

from . import ztyping
from .container import convert_to_container


class ZfitGraphCachable:
    @abstractmethod
    def register_cacher(self, cacher: ZfitGraphCachable):
        raise NotImplementedError

    @abstractmethod
    def add_cache_deps(self, cache_dependents, allow_non_cachable):
        """Add dependents that render the cache invalid if they change.

        Args:
            cache_dependents:
            allow_non_cachable: If `True`, allow `cache_dependents` to be non-cachables.
                If `False`, any `cache_dependents` that is not a `ZfitGraphCachable` will raise an error.

        Raises:
            TypeError: if one of the `cache_dependents` is not a `ZfitGraphCachable` _and_ `allow_non_cachable`
                if `False`.
        """
        pass

    @abstractmethod
    def reset_cache_self(self):
        """Clear the cache of self and all dependent cachers."""
        pass

    @abstractmethod
    def reset_cache(self, reseter):
        pass


class GraphCachable(ZfitGraphCachable):
    graph_caching_methods = []
    instances = weakref.WeakSet()

    def __init__(self, *args, **kwargs):
        self._cache = {}
        self._cachers = weakref.WeakKeyDictionary()
        self.reset_cache_self()
        self.instances.add(self)
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        graph_caching_methods = []
        for func_name in dir(cls):
            if not func_name.startswith("__"):
                func = getattr(cls, func_name)
                if callable(func) and hasattr(func, "zfit_graph_cache_registered"):
                    # assert hasattr(func, "_descriptor_cache"), "TensorFlow internals have changed. Need to update cache"
                    func.zfit_graph_cache_registered = True
                    graph_caching_methods.append(func)
        cls.graph_caching_methods = graph_caching_methods

    def register_cacher(self, cacher: ztyping.CacherOrCachersType):
        """Register a `cacher` that caches values produces by this instance; a dependent.

        Args:
            cacher:
        """
        if not isinstance(cacher, ZfitGraphCachable):
            raise TypeError(f"`cacher` is not a `ZfitGraphCachable` but {type(cacher)}")
        if cacher not in self._cachers:
            self._cachers[cacher] = None

    def add_cache_deps(
        self, cache_deps: ztyping.CacherOrCachersType, allow_non_cachable: bool = True
    ):
        """Add dependencies that render the cache invalid if they change.

        Args:
            cache_deps:
            allow_non_cachable: If `True`, allow `cache_dependents` to be non-cachables.
                If `False`, any `cache_dependents` that is not a `ZfitGraphCachable` will raise an error.

        Raises:
            TypeError: if one of the `cache_dependents` is not a `ZfitGraphCachable` _and_ `allow_non_cachable`
                if `False`.
        """
        cache_deps = convert_to_container(cache_deps)
        for cache_dep in cache_deps:
            if isinstance(cache_dep, ZfitGraphCachable):
                cache_dep.register_cacher(self)
            elif not allow_non_cachable:
                raise TypeError(
                    f"cache_dependent {cache_dep} is not a `ZfitGraphCachable` but {type(cache_dep)}"
                )

    def reset_cache_self(self):
        """Clear the cache of self and all dependent cachers."""
        self._clean_cache()
        self._inform_cachers()

    def reset_cache(self, reseter: ZfitGraphCachable):
        self.reset_cache_self()

    def _clean_cache(self):
        self._cache.clear()
        return

    def _inform_cachers(self):
        for cacher in self._cachers:
            cacher.reset_cache(reseter=self)


def invalidate_graph(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        self = args[0]
        if not isinstance(self, ZfitGraphCachable):
            raise TypeError(
                "Decorator can only be used in a subclass of `ZfitGraphCachable`"
            )

        from .. import run

        if not tf.inside_function():
            run.clear_graph_cache()
        else:
            self.reset_cache(reseter=self)
        # TODO: we could make the whole handling more precise and not just reset the whole graph
        # raise RuntimeError(f"The function {func} is not supported in graph mode as it modifies pieces that invalidate"
        #                    " the graph. If you think this should work, please open an issue on"
        #                    " https://github.com/zfit/zfit/issues/new/choose.")

        return func(*args, **kwargs)

    return wrapped_func


class FunctionCacheHolder(GraphCachable):
    IS_TENSOR = object()

    def __init__(
        self,
        func,
        wrapped_func,
        cachables: ZfitGraphCachable | object | Iterable[ZfitGraphCachable] = None,
        cachables_mapping=None,
        stateless_args: bool | None = None,
        deleter=None,
        keepalive=None,
    ):
        """`tf.function` decorated function holder with caching dependencies on inputs.

        A `tf.function` creates a new graph for every signature that is encountered. It automatically caches them but
        thereby assumes that Python objects are immutable. Any mutation won't be detected. Therefore, an extra wrapper
        is needed. The input signature is compared with firstly checking whether the function is the same and then
        doing an equal comparison of the arguments (maybe too costly?).

        The `FunctionCacheHolder` holds the
         - original python function which serves as the hash of the object
         - wrapped python function, `wrapped_func`
         - the (keyword-)arguments

        If any of the keyword arguments changes in a way that the graph cache is invalid, this holder will have
        `is_valid` set to False and the `wrapped_func` cannot be used anymore, instead a new `tf.function` should
        be created as a call to the `wrapped_func` with the given arguments will result in an outdated graph.

        Args:
            func: Python function that serves as a hash of the holder. Notice that equality is different
                defined.
            wrapped_func: Wrapped `func` with `tf.function`. The holder signals via
                `is_valid` whether this function is still valid to be used.
            cachables: objects that are cached. If they change, the cache is invalidated
            cachables_mapping: keyword arguments to the function. If the values change, the cache is
                invalidated.
            stateless_args: If `True`, the arguments that are normally stateful, such as `tf.Variable`s, are regarded
                as stateless.
        """
        if keepalive is None:
            keepalive = False
        self.deleter = lambda x: None
        self.delete_from_cache = False
        if stateless_args is None:
            stateless_args = False
        self.stateless_args = stateless_args
        if not keepalive:
            wrapped_func = weakref.proxy(wrapped_func, deleter)
        self.wrapped_func = wrapped_func

        self.python_func = func

        if cachables is None and cachables_mapping is None:
            raise ValueError(
                "Both `cachables and `cachables_mapping` are None. One needs to be different from None."
            )
        if cachables is None:
            cachables = []
        if cachables_mapping is None:
            cachables_mapping = {}

        cachables = convert_to_container(cachables, container=list)
        cachables_values = convert_to_container(
            cachables_mapping.values(), container=list
        )
        cachables_all = cachables + cachables_values
        self.immutable_representation = self.create_immutable(
            cachables, cachables_mapping
        )
        self._hash_value = hash((self.python_func, self.immutable_representation))
        super().__init__()  # resets the cache
        self.add_cache_deps(cachables_all)
        self.is_valid = True  # needed to make the cache valid again
        self.deleter = deleter

    def reset_cache_self(self):
        self.is_valid = False
        self.deleter(self)

    def create_immutable(self, args, kwargs):
        """Create a tuple of the args and kwargs by combining them as args + kwargs.keys() + kwargs.values()`

        Args:
            args: list like
            kwargs: dict-like

        Returns:
        """
        # is initialized before the core

        args = list(args)
        kwargs = list(kwargs.keys()) + list(kwargs.values())
        combined = args + kwargs
        combined_cleaned = []
        for obj in combined:
            obj = self.get_immutable_repr_obj(obj)
            combined_cleaned.extend(obj)
        return tuple(combined_cleaned)

    def get_immutable_repr_obj(self, obj):
        from ..core.interfaces import (
            ZfitData,
            ZfitParameter,
            ZfitSpace,
            ZfitLoss,
            ZfitModel,
            ZfitConstraint,
            ZfitPDF,
            ZfitLimit,
        )

        if isinstance(obj, collections.abc.Mapping):
            obj = obj.items()
        if isinstance(
            obj,
            (
                ZfitData,
                ZfitLoss,
                ZfitModel,
                ZfitConstraint,
                ZfitPDF,
                ZfitLimit,
                ZfitSpace,
                uhi.typing.plottable.PlottableHistogram,
            ),
        ):
            obj = self.IS_TENSOR
        elif isinstance(obj, ZfitParameter):
            # if self.stateless_args:
            #     obj = (self.IS_TENSOR,)
            # else:
            obj = (ZfitParameter, obj.name)
        elif tf.is_tensor(obj):
            obj = self.IS_TENSOR
        elif isinstance(obj, (np.ndarray, int, float)):
            obj = self.IS_TENSOR
        elif isinstance(obj, str):
            obj = self.IS_TENSOR
        elif isinstance(obj, collections.abc.Iterable):
            obj_new = []
            for obj_i in obj:
                obj_new.extend(self.get_immutable_repr_obj(obj_i))
            obj = tuple(obj_new)
        else:
            obj = self.IS_TENSOR
        if not isinstance(obj, tuple):
            obj = (obj,)
        return obj

    def __hash__(self) -> int:
        return self._hash_value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FunctionCacheHolder):
            return False
        array_repr_self = self.immutable_representation
        array_repr_other = other.immutable_representation
        try:
            all_ids = all(
                (obj1 is obj2) or (obj1 == obj2)
                for obj1, obj2 in zip_longest(array_repr_self, array_repr_other)
            )
            return all_ids  # TODO: this isn't optimal...
        except ValueError:  # broadcasting does not work
            return False
        except TypeError:  # OperatorNotAllowedError inherits from this
            return False
        # TODO: activate the below? costly, but runs?
        # except OperatorNotAllowedInGraphError:  # we have to assume they're not the same
        #     return False

    def __repr__(self) -> str:
        return f"<FunctionCacheHolder: {self.python_func}, valid={self.is_valid}>"


def clear_graph_cache():
    from zfit.z.zextension import FunctionWrapperRegistry

    for registry in FunctionWrapperRegistry.registries:
        for wrapped_meth in registry.function_cache:
            wrapped_meth = wrapped_meth.wrapped_func
            wrapped_meth._created_variables = None
            wrapped_meth._stateful_fn = None
            wrapped_meth._stateless_fn = None
            wrapped_meth._descriptor_cache.clear()

    for registry in FunctionWrapperRegistry.registries:
        registry.reset()
    for instance in GraphCachable.instances:
        instance.reset_cache("global")
    # Cachable.graph_caching_methods.clear()

    tf.compat.v1.reset_default_graph()



#  Copyright (c) 2023 zfit

from __future__ import annotations

from collections.abc import Iterable

import hist
import zfit_interface as zinterface
from hist.axestuple import NamedAxesTuple

# @tfp.experimental.auto_composite_tensor()
# class Regular(hist.axis.Regular, tfp.experimental.AutoCompositeTensor, family='zfit'):
#     pass
from zfit.core.interfaces import ZfitBinning


class Variable(zinterface.variables.ZfitVar):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name


class SpaceV2:
    def __init__(self, axes):
        self.axes = axes

    def __getitem__(self, key):
        key = to_var_str(key)
        for axis in self.axes:
            if axis.name == key:
                return axis
        else:
            raise KeyError(f"{key} not in {self}.")

    def __iter__(self):
        yield from self.axes

    @property
    def names(self):
        return [axis.name for axis in self]


def to_var_str(value):
    if isinstance(value, str):
        return value
    if isinstance(value, zinterface.variables.ZfitVar):
        return value.name


class Axis(Variable):
    def __init__(self, name):
        super().__init__(name=name)


class UnbinnedAxis(Axis):
    def __init__(self, name, lower=None, upper=None):
        super().__init__(name)
        self.lower = lower
        self.upper = upper


# TODO: fill out below and don't just use the hist objects
class HashableAxisMixin:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.name == "":
            raise ValueError(
                "Currently, a binning has to have a name coinciding with the obs."
            )

    def __hash__(self):
        return hash(tuple(self.edges))


class RegularBinning(HashableAxisMixin, hist.axis.Regular, ZfitBinning, family="zfit"):
    def __init__(self, bins: int, start: float, stop: float, *, name: str) -> None:
        super().__init__(bins, start, stop, name=name, flow=False)


class VariableBinning(
    HashableAxisMixin, hist.axis.Variable, ZfitBinning, family="zfit"
):
    def __init__(self, edges: Iterable[float], *, name: str) -> None:
        super().__init__(edges=edges, name=name, flow=False)


class Binnings(hist.axestuple.NamedAxesTuple):
    pass


def histaxis_to_axis(axis):
    return axis


def axis_to_histaxis(axis):
    return axis


def new_from_axis(axis):
    if isinstance(axis, hist.axis.Regular):
        lower, upper = axis.edges[0], axis.edges[-1]
        return RegularBinning(axis.size, lower, upper, name=axis.name)
    if isinstance(axis, hist.axis.Variable):
        return VariableBinning(axis.edges, name=axis.name)
    raise ValueError(f"{axis} is not a valid axis.")


def histaxes_to_binning(binnings):
    new_binnings = []
    for binning in binnings:
        new_binnings.append(new_from_axis(binning))
    return Binnings(new_binnings)


def binning_to_histaxes(binnings):
    return binnings



#  Copyright (c) 2022 zfit
from __future__ import annotations

import tensorflow as tf

from .base_tf import WrapOptimizer


class Adam(WrapOptimizer):
    # todo see why sphinx does not correctly link to WrapOptimizer for api
    _DEFAULT_name = "Adam"

    def __init__(
        self,
        tol=None,
        learning_rate=0.2,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-08,
        name="Adam",
        **kwargs,
    ):
        # todo write documentation for api.
        optimizer = tf.keras.optimizers.Adam(
            learning_rate=learning_rate,
            beta_1=beta1,
            beta_2=beta2,
            epsilon=epsilon,
            name=name,
        )
        super().__init__(optimizer=optimizer, tol=tol, maxiter=None, **kwargs)



"""Example test for a pdf or function."""

#  Copyright (c) 2024 zfit

import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy.signal
import scipy.stats
import tensorflow as tf

import zfit
import zfit.z.numpy as znp
from zfit import z
from zfit.util.exception import WorkInProgressError

param1_true = 0.3
param2_true = 1.2


class FFTConvPDFV1NoSampling(zfit.pdf.FFTConvPDFV1):
    @zfit.supports()
    def _sample(self, n, limits):
        raise zfit.exception.SpecificFunctionNotImplemented


interpolation_methods = ("linear", "spline", "spline:5", "spline:3")


@pytest.mark.parametrize("interpolation", interpolation_methods)
def test_conv_simple(interpolation):
    n_points = 2432
    obs = zfit.Space("obs1", limits=(-5, 5))
    param1 = zfit.Parameter("param1", -3)
    param2 = zfit.Parameter("param2", 0.3)
    gauss = zfit.pdf.Gauss(0.0, param2, obs=obs)
    func1 = zfit.pdf.Uniform(param1, param2, obs=obs)
    func2 = zfit.pdf.Uniform(-1.2, -1, obs=obs)
    func = zfit.pdf.SumPDF([func1, func2], 0.5)
    conv = zfit.pdf.FFTConvPDFV1(
        func=func, kernel=gauss, n=100, interpolation=interpolation
    )
    if interpolation == "spline:5":
        assert conv._conv_spline_order == 5
    elif interpolation == "spline:3":
        assert conv._conv_spline_order == 3

    x = tf.linspace(-5.0, 5.0, n_points)
    probs = conv.pdf(x=x)

    # true convolution
    true_conv = true_conv_np(
        func, gauss, obs, x, xkernel=tf.linspace(*obs.limit1d, num=n_points)
    )

    integral = conv.integrate(limits=obs)
    probs_np = probs.numpy()
    np.testing.assert_allclose(probs, true_conv, rtol=0.01, atol=0.01)

    assert pytest.approx(1, rel=1e-3) == integral.numpy()
    assert len(probs_np) == n_points

    plt.figure()
    plt.title(f"Conv FFT 1Dim, interpolation={interpolation}")
    plt.plot(x, probs_np, label="zfit")
    plt.plot(x, true_conv, label="numpy")
    plt.legend()
    # pytest.zfit_savefig()


@pytest.mark.parametrize("interpolation", interpolation_methods)
def test_asymetric_limits(interpolation):
    from numpy import linspace

    import zfit
    from zfit.models.convolution import FFTConvPDFV1

    ## Space
    low_obs = -30
    high_obs = 30
    obs = zfit.Space("space", limits=[low_obs, high_obs])

    ## PDFs
    uniform1 = zfit.pdf.Uniform(low=-10, high=10, obs=obs)
    uniform2 = zfit.pdf.Uniform(low=-10, high=10, obs=obs)

    conv_uniforms_1 = FFTConvPDFV1(
        func=uniform1,
        kernel=uniform2,
        limits_kernel=(-18, 18),
        interpolation=interpolation,
    )

    conv_uniforms_2 = FFTConvPDFV1(
        func=uniform1,
        kernel=uniform2,
        limits_kernel=(-12, 12),
        interpolation=interpolation,
    )
    conv_uniforms_3 = FFTConvPDFV1(
        func=uniform1,
        kernel=uniform2,
        limits_kernel=(-25, 12),
        interpolation=interpolation,
    )

    x = linspace(low_obs, high_obs, 300)

    tol = 5e-3
    # If this fails, we're too sensitive
    np.testing.assert_allclose(
        conv_uniforms_1.pdf(x), conv_uniforms_2.pdf(x), rtol=tol, atol=tol
    )

    # this is the "actual" test
    np.testing.assert_allclose(
        conv_uniforms_1.pdf(x), conv_uniforms_3.pdf(x), rtol=tol, atol=tol
    )


@pytest.mark.parametrize("interpolation", interpolation_methods)
def test_conv_1d_shifted(interpolation):
    kerlim = (-3, 3)  # symmetric to make the np conv comparison simple
    obs_kernel = zfit.Space("obs1", limits=kerlim)
    obs = zfit.Space("obs1", limits=(5, 15))
    func1 = zfit.pdf.GaussianKDE1DimV1(obs=obs, data=np.random.uniform(6, 12, size=100))
    # func1 = zfit.pdf.Uniform(6, 12, obs=obs)
    func2 = zfit.pdf.Uniform(11, 11.5, obs=obs)
    func = zfit.pdf.SumPDF([func1, func2], 0.5)

    func1k = zfit.pdf.Gauss(0.0, 1, obs=obs_kernel)
    func2k = zfit.pdf.Gauss(1.0, 0.4, obs=obs_kernel)
    funck = zfit.pdf.SumPDF([func1k, func2k], 0.5)

    conv = zfit.pdf.FFTConvPDFV1(func=func, kernel=funck, n=200)

    xnp = znp.linspace(obs_kernel.rect_lower, obs.rect_upper, 4023)

    # true convolution
    kernel_points = obs_kernel.filter(xnp)
    x = obs.filter(xnp)
    probs = conv.pdf(x=x)
    true_conv = true_conv_np(func, funck, obs, x=x, xkernel=kernel_points)

    integral = conv.integrate(
        limits=obs,
    )
    probs_np = probs.numpy()
    np.testing.assert_allclose(probs_np, true_conv, rtol=0.01, atol=0.01)

    assert pytest.approx(1, rel=1e-3) == integral.numpy()

    plt.figure()
    plt.title("Conv FFT 1Dim shift testing")
    plt.plot(x, probs_np, label="zfit")
    plt.plot(x, true_conv, label="numpy")
    plt.legend()
    pytest.zfit_savefig()


@pytest.mark.parametrize("interpolation", interpolation_methods)
@pytest.mark.flaky(reruns=3)
def test_onedim_sampling(interpolation):
    # there is a sampling shortcut, so we test if it also works without the shortcut
    obs_kernel = zfit.Space("obs1", limits=(-3, 1))
    obs = zfit.Space("obs1", limits=(5, 15))
    func1 = zfit.pdf.Uniform(6, 12, obs=obs)
    func2 = zfit.pdf.Uniform(11, 11.5, obs=obs)
    func = zfit.pdf.SumPDF([func1, func2], 0.5)

    func1k = zfit.pdf.Uniform(-2, 1, obs=obs_kernel)
    func2k = zfit.pdf.Uniform(-0.5, 1.0, obs=obs_kernel)
    funck = zfit.pdf.SumPDF([func1k, func2k], 0.5)
    conv = zfit.pdf.FFTConvPDFV1(
        func=func, kernel=funck, n=200, interpolation=interpolation
    )

    conv_nosample = FFTConvPDFV1NoSampling(
        func=func, kernel=funck, n=200, interpolation=interpolation
    )
    npoints_sample = 10000
    sample = conv.sample(npoints_sample)
    sample_nosample = conv_nosample.sample(npoints_sample)
    x = z.unstack_x(sample)
    xns = z.unstack_x(sample_nosample)
    assert (
        scipy.stats.ks_2samp(x, xns).pvalue > 1e-3
    )  # can vary a lot, but still means close


def true_conv_np(func, gauss1, obs, x, xkernel):
    y_kernel = gauss1.pdf(xkernel)
    y_func = func.pdf(x)
    true_conv = scipy.signal.fftconvolve(y_func, y_kernel, mode="same")
    true_conv /= np.mean(true_conv) * obs.rect_area()
    return true_conv


def true_conv_2d_np(func, gauss1, obsfunc, xfunc, xkernel):
    y_func = func.pdf(xfunc)
    y_kernel = gauss1.pdf(xkernel)
    nfunc = int(np.sqrt(xfunc.shape[0]))
    y_func = tf.reshape(y_func, (nfunc, nfunc))
    nkernel = int(np.sqrt(xkernel.shape[0]))
    y_kernel = tf.reshape(y_kernel, (nkernel, nkernel))
    true_conv = scipy.signal.convolve(y_func, y_kernel, mode="same")
    true_conv /= np.mean(true_conv) * obsfunc.rect_area()
    return tf.reshape(true_conv, xfunc.shape[0])


def test_max_1dim():
    obs1 = zfit.Space("obs1", limits=(-2, 4))
    obs2 = zfit.Space("obs2", limits=(-6, 4))

    param2 = zfit.Parameter("param2", 0.4)
    gauss1 = zfit.pdf.Gauss(1.0, 0.5, obs=obs1)
    gauss22 = zfit.pdf.CrystalBall(0.0, param2, -0.2, 3, obs=obs2)

    obs1func = zfit.Space("obs1", limits=(4, 10))
    obs2func = zfit.Space("obs2", limits=(-6, 4))

    gauss21 = zfit.pdf.Gauss(-0.5, param2, obs=obs2func)
    func1 = zfit.pdf.Uniform(5, 8, obs=obs1func)
    func2 = zfit.pdf.Uniform(6, 7, obs=obs1func)
    func = zfit.pdf.SumPDF([func1, func2], 0.5)
    func = func * gauss21
    gauss = gauss1 * gauss22
    with pytest.raises(WorkInProgressError):
        _ = zfit.pdf.FFTConvPDFV1(func=func, kernel=gauss)


@pytest.mark.skip  # not yet implemented WIP
def test_conv_2D_simple():
    # zfit.run.set_graph_mode(False)  # TODO: remove, just for debugging
    # raise WorkInProgressError("2D convolution not yet implemented, re-activate if so")
    n_points = 1000
    # obs1 = zfit.Space("obs1", limits=(-2, 4))
    # obs2 = zfit.Space("obs2", limits=(-6, 4))
    obs1 = zfit.Space("obs1", limits=(-5, 5))
    obs2 = zfit.Space("obs2", limits=(-5, 5))
    obskernel = obs1 * obs2

    param2 = zfit.Parameter("param2", 0.4)
    gauss1 = zfit.pdf.Gauss(1.0, 0.5, obs=obs1)
    gauss22 = zfit.pdf.CrystalBall(0.0, param2, -0.2, 3, obs=obs2)

    gauss1 = zfit.pdf.Uniform(-1, 1, obs=obs1)
    gauss22 = zfit.pdf.Uniform(-2, 2, obs=obs2)

    obs1func = zfit.Space("obs1", limits=(-10, 10))
    obs2func = zfit.Space("obs2", limits=(-26, 26))
    obs_func = obs1func * obs2func

    gauss21 = zfit.pdf.Gauss(-0.5, param2, obs=obs2func)
    func1 = zfit.pdf.Uniform(2, 8, obs=obs1func)
    func2 = zfit.pdf.Uniform(6, 7, obs=obs1func)
    func = zfit.pdf.SumPDF([func1, func2], 0.5)
    func = func * gauss21
    gauss = gauss1 * gauss22
    conv = zfit.pdf.FFTConvPDFV1(func=func, kernel=gauss)

    start = obs_func.rect_lower
    stop = obs_func.rect_upper
    x_tensor = z.random.uniform((n_points, 2), start, stop)
    x_tensor = tf.reshape(x_tensor, (-1, 2))
    linspace = tf.linspace(start, stop, num=n_points)
    linspace = tf.transpose(tf.meshgrid(*tf.unstack(linspace, axis=-1)))
    linspace_func = tf.reshape(linspace, (-1, 2))

    # linspace_full = tf.linspace((-8, -8), (12, 12), num=n_points)
    # linspace_full = tf.transpose(tf.meshgrid(*tf.unstack(linspace_full, axis=-1)))
    # linspace_full = tf.reshape(linspace_full, (-1, 2))

    linspace_kernel = tf.linspace(
        obskernel.rect_lower, obskernel.rect_upper, num=n_points
    )
    linspace_kernel = tf.transpose(tf.meshgrid(*tf.unstack(linspace_kernel, axis=-1)))
    linspace_kernel = tf.reshape(linspace_kernel, (-1, 2))
    # linspace_kernel = obskernel.filter(linspace_full)
    # linspace_func = obs_func.filter(linspace_full)

    x = zfit.Data.from_tensor(obs=obs_func, tensor=x_tensor)
    linspace_data = zfit.Data.from_tensor(obs=obs_func, tensor=linspace)
    probs_rnd = conv.pdf(x=x)
    probs = conv.pdf(x=linspace_data)

    # Numpy doesn't support ndim convolution?
    true_probs = true_conv_2d_np(
        func, gauss, obsfunc=obs_func, xfunc=linspace_func, xkernel=linspace_kernel
    )
    import matplotlib.pyplot as plt

    # np.testing.assert_allclose(probs, true_probs, rtol=0.2, atol=0.1)
    integral = conv.integrate(
        limits=obs_func,
    )
    assert pytest.approx(1, rel=1e-3) == integral.numpy()
    probs_np = probs_rnd.numpy()
    assert len(probs_np) == n_points
    # probs_plot = np.reshape(probs_np, (-1, n_points))
    # x_plot = linspace[0:, ]
    # probs_plot_projx = np.sum(probs_plot, axis=0)
    # plt.plot(x_plot, probs_np)
    # probs_plot = np.reshape(probs_np, (n_points, n_points))
    # plt.imshow(probs_plot)
    # plt.show()

    true_probsr = tf.reshape(true_probs, (n_points, n_points))
    probsr = tf.reshape(probs, (n_points, n_points))
    plt.figure()
    plt.imshow(true_probsr, label="true probs")
    plt.title("true probs")

    plt.figure()
    plt.imshow(probsr, label="zfit conv")
    plt.title("zfit conv")

    # test the sampling
    conv_nosample = FFTConvPDFV1NoSampling(func=func, kernel=gauss)

    npoints_sample = 10000
    sample = conv.sample(npoints_sample)
    sample_nosample = conv_nosample.sample(npoints_sample)
    x, y = z.unstack_x(sample)
    xns, yns = z.unstack_x(sample_nosample)

    plt.figure()
    plt.title("FFT conv, custom sampling, addition")
    plt.hist2d(x, y, bins=30)
    # pytest.zfit_savefig()

    plt.figure()
    plt.title("FFT conv, fallback sampling, accept-reject")
    plt.hist2d(xns, yns, bins=30)
    # pytest.zfit_savefig()

    plt.figure()
    plt.title("FFT conv x projection")
    plt.hist(x.numpy(), bins=50, label="custom", alpha=0.5)
    plt.hist(xns.numpy(), bins=50, label="fallback", alpha=0.5)
    plt.legend()
    # pytest.zfit_savefig()

    plt.figure()
    plt.title("FFT conv y projection")
    plt.hist(y.numpy(), bins=50, label="custom", alpha=0.5)
    plt.hist(yns.numpy(), bins=50, label="fallback", alpha=0.5)
    plt.legend()
    # pytest.zfit_savefig()



"""Module for testing of the zfit components.

Contains a singleton instance to register new PDFs and let them be tested.
"""

#  Copyright (c) 2024 zfit

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Iterable, Callable

import scipy.stats

from .interfaces import ZfitPDF
from ..util.container import convert_to_container

__all__ = ["tester"]

import scipy.integrate


def check_integrate(func, limits, norm_range):
    if norm_range is not False:
        return check_integrate(func, limits, False) / check_integrate(
            func, norm_range, False
        )
    lower, upper = limits.limid1d
    return scipy.integrate.quad(func, lower, upper)


class AutoTester:
    def __init__(self):
        self.pdfs = []

    def register_pdf(
        self,
        pdf_class: ZfitPDF,
        params_factories: Callable | Iterable[Callable],
        scipy_dist: scipy.stats.rv_continuous = None,
        analytic_int_axes: None | int | list[tuple[int, ...]] = None,
    ):
        # if not isinstance(pdf_class, ZfitPDF):
        #     raise TypeError(f"PDF {pdf_class} is not a ZfitPDF.")
        params_factories = convert_to_container(params_factories)

        if isinstance(analytic_int_axes, tuple):
            raise TypeError(
                "`analytic_int_axes` is either a number or a list of tuples."
            )
        analytic_int_axes = convert_to_container(
            analytic_int_axes, non_containers=[tuple]
        )
        if analytic_int_axes is not None:
            isinstance(analytic_int_axes)
        registration = OrderedDict(
            (
                ("pdf_class", pdf_class),
                ("params", params_factories),
                ("scipy_dist", scipy_dist),
            )
        )
        self.pdfs.append(registration)

    def create_parameterized_pdfs(self):
        if len(self.pdfs) == 0:
            return [], []

        argnames = list(self.pdfs[0].keys())
        argvals = [[] * len(argnames)]

        for pdf in self.pdfs:
            for i, param in enumerate(pdf.values()):
                argvals[i].append(param)

        return argnames, argvals


tester = AutoTester()



"""Module is a backend which consists of two parts, a control layer and a numpy-like layer.

z itself wraps (or will in the future) control flow such as ``cond``, ``while_loop`` etc. that can be found in
different libraries supporting compiled, vectorized computation such as JAX and TensorFlow.

Numpy is a Numpy like interface for math functions and arrays.

This module is intended to replace tensorflow specific methods and datastructures with equivalent or similar versions
in the numpy api. This should help make zfit as a project portable to alternatives of tensorflow should it be
necessary in the future.
At the moment it is simply an alias for the numpy api of tensorflow.
See https://www.tensorflow.org/guide/tf_numpy for more a guide to numpy api in tensorflow.
See https://www.tensorflow.org/api_docs/python/tf/experimental/numpy for the complete numpy api in tensorflow.

Recommended way of importing:
>>> Import zfit.z.numpy as znp
"""

#  Copyright (c) 2022 zfit

# TODO: dymamic imports?
# import tensorflow.experimental.numpy as _tnp  # this way we do get the autocompletion
# numpy = _tnp  # for static code analysis
# import sys
# sys.modules['zfit.z.numpy'] = _tnp


from . import math, random, unstable
from .tools import _get_ndims
from .wrapping_tf import (
    check_numerics,
    complex,
    convert_to_tensor,
    exp,
    pow,
    random_normal,
    random_uniform,
    reduce_prod,
    reduce_sum,
    sqrt,
    square,
)
from .zextension import (
    abs_square,
    constant,
    function as function,
    function_sampling,
    function_tf_input,
    nth_pow,
    pi,
    py_function,
    run_no_nan,
    safe_where,
    stack_x,
    to_complex,
    to_real,
    unstack_x,
)

# numpy = _tnp



#  Copyright (c) 2023 zfit
import numpy as np
import pytest

import zfit
from zfit.core.binnedpdf import binned_rect_integration


def test_calculate_scaled_edges(edges_bins1):
    from zfit.core.binnedpdf import cut_edges_and_bins

    edges, true_scaled_edges, limits, limits_true, value_scaling, values = edges_bins1
    scaled_edges, bins, _ = cut_edges_and_bins(edges, limits)
    np.testing.assert_allclose(true_scaled_edges[0], scaled_edges[0])
    np.testing.assert_allclose(true_scaled_edges[1], scaled_edges[1])
    np.testing.assert_allclose(true_scaled_edges[2], scaled_edges[2])


def test_binned_rect_integration(edges_bins1):
    edges, true_scaled_edges, limits, limits_true, value_scaling, values = edges_bins1

    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = limits_true.area() * value_scaling
    assert pytest.approx(float(true_integral)) == float(integral)

    # integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    # true_integral = value_scaling * np.prod([e.shape.num_elements()
    #                                                               for e in true_scaled_edges])
    # assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_simple():
    import zfit.z.numpy as znp

    scaling = 0.2
    values = znp.ones((2, 2)) * scaling
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (0, 5))
    lim2 = zfit.Space("b", (1, 9))
    limits = lim1 * lim2
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = 5 * 8 * scaling  # area lim1, area lim2, scaling
    assert pytest.approx(float(true_integral)) == float(integral)

    integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    true_integral = 4 * scaling  # 4 elements

    assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_simple_too_large():
    import zfit.z.numpy as znp

    scaling = 0.2
    values = znp.ones((2, 2)) * scaling
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (-1, 10))
    lim2 = zfit.Space("b", (0.5, 19))
    limits = lim1 * lim2
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = 5 * 8 * scaling  # area lim1, area lim2, scaling
    assert pytest.approx(float(true_integral)) == float(integral)

    integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    true_integral = 4 * scaling  # 4 elements

    assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_simple_scaled():
    import zfit.z.numpy as znp

    scaling = 0.2
    values = znp.ones((2, 2)) * scaling
    reducefac = 0.8  # reduce the limits by a factor of 0.8
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (0.5, 4.5))
    lim2 = zfit.Space("b", (1.8, 8.2))
    limits = lim1 * lim2
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = 5 * 8 * scaling * reducefac**2  # area lim1, area lim2, scaling
    assert pytest.approx(float(true_integral)) == float(integral)

    integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    true_integral = (2 * reducefac) ** 2 * scaling  # 4 elements

    assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_simple_scaled_asym():
    import zfit.z.numpy as znp

    scaling = 0.2
    values = znp.ones((2, 2)) * scaling
    reducefac = 0.85  # reduce the limits by a factor of 0.8
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (0.25, 4.5))
    lim2 = zfit.Space("b", (1.4, 8.2))
    limits = lim1 * lim2
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = 5 * 8 * scaling * reducefac**2  # area lim1, area lim2, scaling
    assert pytest.approx(float(true_integral)) == float(integral)

    integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    true_integral = (2 * reducefac) ** 2 * scaling  # 4 elements

    assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_scaled_asym():
    import zfit.z.numpy as znp

    values = znp.array([[1.0, 3.0], [2.0, 4.0]])
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (1.2, 4.7))
    lim2 = zfit.Space("b", (2.1, 8.2))
    limits = lim1 * lim2
    binw11 = 1.3
    binw21 = 2.9
    binw12 = 2.2
    binw22 = 3.2

    # using densities
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    true_integral = (
        1 * binw11 * binw21
        + 2 * binw12 * binw21
        + 3 * binw11 * binw22
        + 4 * binw12 * binw22
    )
    assert pytest.approx(float(true_integral)) == float(integral)

    # using counts
    integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    true_integral = (
        1 * binw11 * binw21
        + 2 * binw12 * binw21
        + 3 * binw11 * binw22
        + 4 * binw12 * binw22
    )
    true_integral /= 10  # each bin has an area of 10
    assert pytest.approx(float(true_integral)) == float(integral)


def test_binned_partial_scaled_asym_axis0():
    import zfit.z.numpy as znp

    values = znp.array([[1.0, 3.0], [2.0, 4.0]])
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (1.2, 4.7))
    integral = binned_rect_integration(density=values, edges=edges, limits=lim1, axis=0)
    binw11 = 1.3
    binw21 = 4
    binw12 = 2.2
    binw22 = 4
    true_integral = (
        1 * binw11 * binw21 + 2 * binw12 * binw21,
        3 * binw11 * binw22 + 4 * binw12 * binw22,
    )
    assert pytest.approx(true_integral) == zfit.run(integral)


def test_binned_partial_scaled_asym_axis1():
    import zfit.z.numpy as znp

    values = znp.array([[1.0, 3.0], [2.0, 4.0]])
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim2 = zfit.Space("b", (2.1, 8.2))

    integral = binned_rect_integration(density=values, edges=edges, limits=lim2, axis=1)
    binw11 = 2.5
    binw21 = 2.9
    binw12 = 2.5
    binw22 = 3.2
    true_integral = (
        1 * binw11 * binw21 + 3 * binw11 * binw22,
        2 * binw12 * binw21 + 4 * binw12 * binw22,
    )
    assert pytest.approx(true_integral) == zfit.run(integral)


def test_binned_scaled_asym_one():
    import zfit.z.numpy as znp

    values = znp.array([[1.0, 3.0], [2.0, 4.0]])
    edges = [[0.0, 2.5, 5.0], [1.0, 5.0, 9.0]]
    lim1 = zfit.Space("a", (0, 4.7))
    lim2 = zfit.Space("b", (1, 9))
    limits = lim1 * lim2
    integral = binned_rect_integration(density=values, edges=edges, limits=limits)
    binw11 = 2.5
    binw21 = 4
    binw12 = 2.2
    binw22 = 4
    true_integral = (
        1 * binw11 * binw21
        + 2 * binw12 * binw21
        + 3 * binw11 * binw22
        + 4 * binw12 * binw22
    )
    assert pytest.approx(float(true_integral)) == float(integral)

    # integral = binned_rect_integration(counts=values, edges=edges, limits=limits)
    # true_integral = (2 * reducefac) ** 2 * scaling  # 4 e


def test_partial_binned_rect_integration(edges_bins1):
    edges, true_scaled_edges, limits, limits_true, value_scaling, values = edges_bins1
    limits_part = limits.with_obs(["a", "c"])
    limits_nonint = limits.with_obs("b")
    limits_nonint = limits_nonint.with_limits([1.9, 4.2])
    limits_full = (limits_part * limits_nonint).with_obs(["a", "b", "c"])
    integral = binned_rect_integration(
        density=values, edges=edges, limits=limits_part, axis=[0, 2]
    )
    assert integral.shape[0] == edges[1].shape[1] - 1
    full_integral = binned_rect_integration(
        density=values, edges=edges, limits=limits_full
    )
    assert pytest.approx(np.sum(integral)) == float(full_integral)


@pytest.fixture()
def edges_bins1():
    import zfit
    import zfit.z.numpy as znp

    lower3 = 1.5
    upper3 = 2.5
    num1 = 6
    num2 = 10
    num3 = 12
    value_scaling = 5.2
    values = znp.ones(shape=(num1 - 1, num2 - 1, num3 - 1)) * value_scaling
    edges3 = znp.linspace(lower3, upper3, num=num3)[None, None, ...]
    edges = [
        znp.linspace(0, 5, num=num1)[..., None, None],
        znp.linspace(2, 4, num=num2)[None, ..., None],
        edges3,
    ]
    lower1 = 1.5
    upper1 = 4.1
    lower2 = 2.5
    upper2 = 4.2
    true_scaled_edges = [
        znp.array([lower1, 2, 3, 4, upper1])[..., None, None],
        znp.array(
            [
                lower2,
                2.66666667,
                2.88888889,
                3.11111111,
                3.33333333,
                3.55555556,
                3.77777778,
                4.0,
            ]
        )[None, ..., None],
        edges3,
    ]
    limits = (
        zfit.Space(obs=["a"], limits=[lower1, upper1])
        * zfit.Space(["b"], [lower2, upper2])
        * zfit.Space(["c"], [lower3, upper3])
    )

    limits_true = (
        zfit.Space(obs=["a"], limits=[lower1, upper1])
        * zfit.Space(["b"], [lower2, 4])
        * zfit.Space(["c"], [lower3, upper3])
    )

    return edges, true_scaled_edges, limits, limits_true, value_scaling, values



#  Copyright (c) 2022 zfit
from __future__ import annotations

import tensorflow_probability as tfp
import zfit_interface.variables

import zfit.util.container


@tfp.experimental.auto_composite_tensor()
class VarSupports(tfp.experimental.AutoCompositeTensor):
    def __init__(
        self,
        var,
        *,
        full=None,
        space=None,
        scalar=None,
        vectorspace=None,
        binned=None,
        data=None,
        types=None,
    ):
        types = zfit.util.container.convert_to_container(types)
        if types:
            if full or space or binned or data or scalar or vectorspace:
                raise ValueError
        elif full:
            if space or binned or data or types or scalar or vectorspace:
                raise ValueError
        elif not (space or scalar or vectorspace or binned or data):
            raise ValueError("Need to support at least something.")
        if data:
            scalar = True
        if vectorspace:
            space = True

        if not isinstance(var, zfit_interface.variables.ZfitVar):
            raise TypeError(f"var has to be a ZfitVar, not {var}.")
        self.var = var
        self.full = full or False
        self.scalar = scalar or False
        self.vectorspace = vectorspace or False
        self.space = space or False
        self.binned = binned or False
        self.data = data or False
        self.types = types or []



#  Copyright (c) 2023 zfit
"""..

todo:: Add module docstring.
"""
from .core.serialmixin import NumpyArrayNotSerializableError
from .util.checks import NONE
from .util.deprecation import deprecated
from .util.exception import (
    AnalyticIntegralNotImplemented,
    AnalyticNotImplemented,
    AnalyticSamplingNotImplemented,
    FunctionNotImplemented,
    IllegalInGraphModeError,
    InitNotImplemented,
    MaximumIterationReached,
    MinimizerSubclassingError,
    MultipleLimitsNotImplemented,
    NameAlreadyTakenError,
    NormRangeNotImplemented,
    SpecificFunctionNotImplemented,
    VectorizedLimitsNotImplemented,
    ShapeIncompatibleError,
    LogicalUndefinedOperationError,
)

__all__ = [
    "NameAlreadyTakenError",
    "IllegalInGraphModeError",
    "NormRangeNotImplemented",
    "MultipleLimitsNotImplemented",
    "VectorizedLimitsNotImplemented",
    "AnalyticNotImplemented",
    "FunctionNotImplemented",
    "AnalyticSamplingNotImplemented",
    "AnalyticIntegralNotImplemented",
    "SpecificFunctionNotImplemented",
    "MinimizerSubclassingError",
    "MaximumIterationReached",
    "ShapeIncompatibleError",
    "InitNotImplemented",
    "LogicalUndefinedOperationError",
    "NumpyArrayNotSerializableError",
    "NONE",
]


class VectorizedLimitsNotImplementedError(VectorizedLimitsNotImplemented):
    @deprecated(None, "Use VectorizedLimitsNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class SpecificFunctionNotImplementedError(SpecificFunctionNotImplemented):
    @deprecated(None, "Use SpecificFunctionNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NormRangeNotImplementedError(NormRangeNotImplemented):
    @deprecated(None, "Use NormRangeNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class MultipleLimitsNotImplementedError(MultipleLimitsNotImplemented):
    @deprecated(None, "Use MultipleLimitsNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class FunctionNotImplementedError(FunctionNotImplemented):
    @deprecated(None, "Use FunctionNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticSamplingNotImplementedError(AnalyticSamplingNotImplemented):
    @deprecated(None, "Use AnalyticSamplingNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticIntegralNotImplementedError(AnalyticIntegralNotImplemented):
    @deprecated(None, "Use AnalyticIntegralNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class AnalyticNotImplementedError(AnalyticNotImplemented):
    @deprecated(None, "Use AnalyticNotImplemented instead.")
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



#  Copyright (c) 2022 zfit

__all__ = [
    "BasePDF",
    "BaseFunctor",
    "Exponential",
    "CrystalBall",
    "DoubleCB",
    "Gauss",
    "Uniform",
    "TruncatedGauss",
    "WrapDistribution",
    "Cauchy",
    "Poisson",
    "Chebyshev",
    "Legendre",
    "Chebyshev2",
    "Hermite",
    "Laguerre",
    "RecursivePolynomial",
    "ProductPDF",
    "SumPDF",
    "GaussianKDE1DimV1",
    "KDE1DimGrid",
    "KDE1DimExact",
    "KDE1DimFFT",
    "KDE1DimISJ",
    "FFTConvPDFV1",
    "ConditionalPDFV1",
    "ZPDF",
    "SimplePDF",
    "SimpleFunctorPDF",
    "UnbinnedFromBinnedPDF",
    "BinnedFromUnbinnedPDF",
    "HistogramPDF",
    "SplineMorphingPDF",
    "BinwiseScaleModifier",
    "BinnedSumPDF",
    "SplinePDF",
]

from .core.basepdf import BasePDF
from .models.basic import Exponential
from .models.binned_functor import BinnedSumPDF
from .models.conditional import ConditionalPDFV1
from .models.convolution import FFTConvPDFV1
from .models.dist_tfp import (
    Cauchy,
    Gauss,
    Poisson,
    TruncatedGauss,
    Uniform,
    WrapDistribution,
)
from .models.functor import BaseFunctor, ProductPDF, SumPDF
from .models.histmodifier import BinwiseScaleModifier
from .models.histogram import HistogramPDF
from .models.interpolation import SplinePDF
from .models.kde import (
    GaussianKDE1DimV1,
    KDE1DimExact,
    KDE1DimFFT,
    KDE1DimGrid,
    KDE1DimISJ,
)
from .models.morphing import SplineMorphingPDF
from .models.physics import CrystalBall, DoubleCB
from .models.polynomials import (
    Chebyshev,
    Chebyshev2,
    Hermite,
    Laguerre,
    Legendre,
    RecursivePolynomial,
)
from .models.special import ZPDF, SimpleFunctorPDF, SimplePDF
from .models.tobinned import BinnedFromUnbinnedPDF
from .models.unbinnedpdf import UnbinnedFromBinnedPDF



#  Copyright (c) 2023 zfit
from __future__ import annotations

import functools

import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops

from zfit.core import interfaces as zinterfaces


def register_tensor_conversion(
    convertable, name=None, overload_operators=True, priority=10
):  # higher than any tf conversion
    def _dense_var_to_tensor(var, dtype=None, name=None, as_ref=False):
        return var._dense_var_to_tensor(dtype=dtype, name=name, as_ref=as_ref)

    ops.register_tensor_conversion_function(
        convertable, _dense_var_to_tensor, priority=priority
    )
    if name:
        pass
        # _pywrap_utils.RegisterType(name, convertable)

    if overload_operators:
        convertable._OverloadAllOperators(cls=convertable)


class OverloadableMixin:
    # Conversion to tensor.
    @staticmethod
    def _TensorConversionFunction(
        v, dtype=None, name=None, as_ref=False
    ):  # pylint: disable=invalid-name
        """Utility function for converting a Variable to a Tensor."""
        _ = name
        if dtype and not dtype.is_compatible_with(v.dtype):
            raise ValueError(
                "Incompatible type conversion requested to type '%s' for variable "
                "of type '%s'" % (dtype.name, v.dtype.name)
            )
        if as_ref:
            return v._ref()  # pylint: disable=protected-access
        else:
            return v.value()

    def _dense_var_to_tensor(self, dtype=None, name=None, as_ref=False):
        del name
        if dtype and not dtype.is_compatible_with(self.dtype):
            raise ValueError(
                "Incompatible type conversion requested to type '%s' for variable "
                "of type '%s'" % (dtype.name, self.dtype.name)
            )
        if as_ref:
            if hasattr(self, "_ref"):
                return self._ref()
            else:
                raise RuntimeError("Why is this needed?")
        else:
            return self.value()

    def _AsTensor(self):
        return self.value()

    @staticmethod
    def _OverloadAllOperators(cls):  # pylint: disable=invalid-name
        """Register overloads for all operators."""
        for operator in tf.Tensor.OVERLOADABLE_OPERATORS:
            cls._OverloadOperator(cls, operator)
        # For slicing, bind getitem differently than a tensor (use SliceHelperVar
        # instead)
        # pylint: disable=protected-access
        setattr(cls, "__getitem__", array_ops._SliceHelperVar)

    @staticmethod
    def _OverloadOperator(cls, operator):  # pylint: disable=invalid-name
        """Defer an operator overload to ``ops.Tensor``.

        We pull the operator out of ops.Tensor dynamically to avoid ordering issues.
        Args:
          operator: string. The operator name.
        """
        # We can't use the overload mechanism on __eq__ & __ne__ since __eq__ is
        # called when adding a variable to sets. As a result we call a.value() which
        # causes infinite recursion when operating within a GradientTape
        # TODO(gjn): Consider removing this
        if operator == "__eq__" or operator == "__ne__":
            return

        tensor_oper = getattr(tf.Tensor, operator)

        def _run_op(a, *args, **kwargs):
            # pylint: disable=protected-access
            return tensor_oper(a.value(), *args, **kwargs)

        functools.update_wrapper(_run_op, tensor_oper)
        setattr(cls, operator, _run_op)


class OverloadableMixinValues:
    # Conversion to tensor.
    @staticmethod
    def _TensorConversionFunction(
        v, dtype=None, name=None, as_ref=False
    ):  # pylint: disable=invalid-name
        """Utility function for converting a Variable to a Tensor."""
        _ = name
        if dtype and not dtype.is_compatible_with(v.dtype):
            raise ValueError(
                "Incompatible type conversion requested to type '%s' for variable "
                "of type '%s'" % (dtype.name, v.dtype.name)
            )
        if as_ref:
            return v._ref()  # pylint: disable=protected-access
        else:
            return v.values()

    def _dense_var_to_tensor(self, dtype=None, name=None, as_ref=False):
        del name
        if dtype and not dtype.is_compatible_with(self.dtype):
            raise ValueError(
                "Incompatible type conversion requested to type '%s' for variable "
                "of type '%s'" % (dtype.name, self.dtype.name)
            )
        if as_ref:
            if hasattr(self, "_ref"):
                return self._ref()
            else:
                raise RuntimeError("Why is this needed?")
        else:
            return self.values()

    def _AsTensor(self):
        return self.values()

    @staticmethod
    def _OverloadAllOperators(cls):  # pylint: disable=invalid-name
        """Register overloads for all operators."""
        for operator in tf.Tensor.OVERLOADABLE_OPERATORS:
            cls._OverloadOperator(cls, operator)
        # For slicing, bind getitem differently than a tensor (use SliceHelperVar
        # instead)
        # pylint: disable=protected-access
        setattr(cls, "__getitem__", array_ops._SliceHelperVar)

    @staticmethod
    def _OverloadOperator(cls, operator):  # pylint: disable=invalid-name
        """Defer an operator overload to ``ops.Tensor``.

        We pull the operator out of ops.Tensor dynamically to avoid ordering issues.
        Args:
          operator: string. The operator name.
        """
        # We can't use the overload mechanism on __eq__ & __ne__ since __eq__ is
        # called when adding a variable to sets. As a result we call a.value() which
        # causes infinite recursion when operating within a GradientTape
        # TODO(gjn): Consider removing this
        if operator == "__eq__" or operator == "__ne__":
            return

        tensor_oper = getattr(tf.Tensor, operator)

        def _run_op(a, *args, **kwargs):
            # pylint: disable=protected-access
            return tensor_oper(a.values(), *args, **kwargs)

        functools.update_wrapper(_run_op, tensor_oper)
        setattr(cls, operator, _run_op)


class MetaBaseParameter(
    type(tf.Variable), type(zinterfaces.ZfitParameter)
):  # resolve metaclasses
    pass



#  Copyright (c) 2024 zfit
import pickle
import platform
import sys

import numpy as np
import pytest

import zfit
from zfit import z
from zfit.minimizers.errors import compute_errors
from zfit.minimizers.fitresult import FitResult

true_a = 2.4
true_b = 1.1
true_c = -0.3

true_val = [true_a, true_b, true_c]
true_ntot = 15000
true_yields = [0.8 * true_ntot, 0.2 * true_ntot]


def minimizer_ids(minimizer_class_and_kwargs):
    return minimizer_class_and_kwargs[0].__name__.split(".")[-1]


def create_loss(n=15000, weights=None, extended=None, constraints=None):
    avalue = 1.5
    a_param = zfit.Parameter("variable_a15151", avalue, -1.0, 20.0, step_size=0.1)
    a_param.init_val = avalue
    bvalue = 1.9
    b_param = zfit.Parameter("variable_b15151", bvalue, 0, 20)
    b_param.init_val = bvalue
    cvalue = -0.03
    c_param = zfit.Parameter("variable_c15151", cvalue, -1, 0.0)
    c_param.init_val = cvalue
    obs1 = zfit.Space(obs="obs1", limits=(-2.4, 9.1))

    # load params for sampling
    a_param.set_value(true_a)
    b_param.set_value(true_b)
    c_param.set_value(true_c)

    if extended:
        yieldgauss_init = n * 0.85
        yieldgauss = (
            zfit.Parameter("yieldgauss", yieldgauss_init, -100, n * 2)
            if extended
            else None
        )
        yieldgauss.init_val = yieldgauss_init
        yieldexp_init = n * 0.23
        yieldexp = (
            zfit.Parameter("yieldepx", yieldexp_init, -100, n) if extended else None
        )
        yieldexp.init_val = yieldexp_init
    else:
        yieldgauss = None
        yieldexp = None
    gauss1 = zfit.pdf.Gauss(mu=a_param, sigma=b_param, obs=obs1, extended=yieldgauss)
    exp1 = zfit.pdf.Exponential(lam=c_param, obs=obs1, extended=yieldexp)

    fracs = None if extended else 0.2
    sum_pdf1 = zfit.pdf.SumPDF((gauss1, exp1), fracs=fracs)

    if extended:
        yieldgauss.set_value(true_yields[0])
        yieldexp.set_value(true_yields[1])

    sampled_data = sum_pdf1.create_sampler(n=n)
    sampled_data.resample()

    if weights is not None:
        sampled_data.set_weights(weights)

    constraint = (
        zfit.constraint.GaussianConstraint(a_param, true_a, 0.1)
        if constraints
        else None
    )
    Loss = zfit.loss.ExtendedUnbinnedNLL if extended else zfit.loss.UnbinnedNLL
    loss = Loss(model=sum_pdf1, data=sampled_data, constraints=constraint)

    if extended:
        params = (a_param, b_param, c_param, yieldgauss, yieldexp)
    else:
        params = (a_param, b_param, c_param)
    return loss, params


def create_fitresult(
    minimizer_class_and_kwargs, n=15000, weights=None, extended=None, constraints=None
):
    loss, all_params = create_loss(
        n=n, weights=weights, extended=extended, constraints=constraints
    )

    true_minimum = loss.value(full=False).numpy()

    if extended:
        a_param, b_param, c_param, yieldgauss, yieldexp = all_params
    else:
        a_param, b_param, c_param = all_params

    minimizer_class, minimizer_kwargs, test_error = minimizer_class_and_kwargs
    minimizer = minimizer_class(**minimizer_kwargs)

    for param in all_params:
        param.assign(param.init_val)  # reset the value
    for ntry in range(3):
        result = minimizer.minimize(loss=loss)
        cur_val = loss.value(full=False).numpy()
        if result.valid:
            break
        else:  # vary param.init_val slightly
            for param in all_params:
                param.assign(param.init_val + np.random.normal(scale=0.1))
    else:
        assert (
            False
        ), "Tried to minimize but failed 3 times, this is treated as an error."
    assert cur_val < true_minimum + 0.1, "Fit did not converge to true minimum"
    aval, bval, cval = (
        result.params[p]["value"] for p in all_params[:3]
    )  # not including yields

    ret = {
        "result": result,
        "true_min": true_minimum,
        "cur_val": cur_val,
        "a": aval,
        "b": bval,
        "c": cval,
        "a_param": a_param,
        "b_param": b_param,
        "c_param": c_param,
    }
    if extended:
        ret["yieldgauss"] = yieldgauss
        ret["yieldexp"] = yieldexp
        ret["ngauss"] = result.params[yieldgauss]["value"]
        ret["nexp"] = result.params[yieldexp]["value"]

    return ret


@pytest.mark.parametrize("do_pickle", [True, False], ids=["pickle", "no_pickle"])
@pytest.mark.parametrize("weights", [None, np.random.normal(1, 0.1, true_ntot)])
@pytest.mark.parametrize("extended", [True, False], ids=["extended", "not_extended"])
def test_set_values_fitresult(do_pickle, weights, extended):
    upper1 = 5.33
    lower1 = 0.0
    param1 = zfit.Parameter("param1", 2.0, lower1, upper1)
    param1.set_value(lower1)
    assert pytest.approx(zfit.run(param1.value())) == lower1
    param1.set_value(upper1)
    assert pytest.approx(zfit.run(param1.value())) == upper1
    with pytest.raises(ValueError):
        param1.set_value(lower1 - 0.001)
    with pytest.raises(ValueError):
        param1.set_value(upper1 + 0.001)

    fitresult = create_fitresult(
        (zfit.minimize.Minuit, {}, True), weights=weights, extended=extended
    )
    result = fitresult["result"]
    param_b = fitresult["b_param"]
    param_c = fitresult["c_param"]

    val_b = fitresult["b"]
    val_c = fitresult["c"]
    with pytest.raises(ValueError):
        param_b.set_value(999)
    param_c.assign(9999)
    if do_pickle:
        result.freeze()
        result = pickle.loads(pickle.dumps(result))
    with zfit.param.set_values([param_c, param_b], values=result):
        assert zfit.run(param_b.value()) == val_b
        assert zfit.run(param_c.value()) == val_c

    # test partial
    param_new = zfit.Parameter("param_new", 42.0, 12.0, 48.0)
    with pytest.raises(ValueError):
        zfit.param.set_values(
            [param_c, param_new], values=result
        )  # allow_partial by default false

    zfit.param.set_values(
        [param_c, param_new], values=result, allow_partial=True
    )  # allow_partial by default false
    assert zfit.run(param_c.value()) == val_c

    # test partial in case we have nothing to set
    param_d = zfit.Parameter("param_d", 12)
    with pytest.raises(ValueError):
        zfit.param.set_values([param_d], result)
    zfit.param.set_values([param_d], result, allow_partial=True)


minimizers = [
    (zfit.minimize.ScipyTrustConstrV1, {}, True),
    (zfit.minimize.Minuit, {}, True),
    (zfit.minimize.NLoptLBFGSV1, {}, True),
]

if not platform.system() in (
    "Darwin",
    "Windows",
):  # TODO: Ipyopt installation on macosx not working
    minimizers.append((zfit.minimize.IpyoptV1, {}, False))
# sort for xdist: https://github.com/pytest-dev/pytest-xdist/issues/432
minimizers = sorted(minimizers, key=lambda val: repr(val))


@pytest.mark.parametrize(
    "minimizer_class_and_kwargs", minimizers, ids=lambda val: val[0].__name__
)
@pytest.mark.parametrize("weights", [np.random.normal(1, 0.1, true_ntot), None])
@pytest.mark.parametrize("extended", [True, False], ids=["extended", "not_extended"])
def test_freeze(minimizer_class_and_kwargs, weights, extended):
    result = create_fitresult(
        minimizer_class_and_kwargs, weights=weights, extended=extended
    )["result"]

    try:
        pickle.dumps(result)
    except Exception:
        pass
    result.covariance()
    result.errors()
    result.hesse()
    result.freeze()

    dumped = pickle.dumps(result)
    loaded = pickle.loads(dumped)
    test = loaded
    true = result
    assert test.fmin == true.fmin
    assert test.edm == true.edm

    for testval, trueval in zip(test.params.values(), true.params.values()):
        assert testval == trueval

    assert test.valid == true.valid
    assert test.status == true.status
    assert test.message == true.message
    assert test.converged == true.converged
    assert test.params_at_limit == true.params_at_limit


@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers, ids=minimizer_ids)
def test_fmin(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results["result"]
    assert pytest.approx(results["cur_val"]) == result.fmin


@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers, ids=minimizer_ids)
def test_params(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results["result"]
    np.testing.assert_allclose(true_val, result.values, rtol=0.2)
    for param in result.params:
        assert result.values[param] == result.params[param]["value"]
        assert result.values[param.name] == result.params[param]["value"]
        assert result.values[param.name] == result.params[param.name]["value"]


def test_params_at_limit():
    loss, (param_a, param_b, param_c) = create_loss(n=5000)
    old_lower = param_a.lower
    param_a.lower = param_a.upper
    param_a.assign(param_a.upper + 5)
    minimizer = zfit.minimize.Minuit(gradient=True, tol=10.0)
    result = minimizer.minimize(loss)
    param_a.assign(100)
    assert param_a.at_limit
    assert result.params_at_limit
    param_a.lower = old_lower
    assert param_a.at_limit
    assert not result.valid


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers, ids=minimizer_ids)
@pytest.mark.parametrize("use_weights", [False, True], ids=["no_weights", "weights"])
@pytest.mark.parametrize("extended", [True, False], ids=["extended", "not_extended"])
def test_covariance(minimizer_class_and_kwargs, use_weights, extended):
    n = true_ntot
    if use_weights:
        weights = np.random.normal(1, 0.001, n)
    else:
        weights = None

    results = create_fitresult(
        minimizer_class_and_kwargs=minimizer_class_and_kwargs,
        n=n,
        weights=weights,
        extended=extended,
    )
    result = results["result"]
    hesse = result.hesse()
    a = results["a_param"]
    b = results["b_param"]
    c = results["c_param"]

    with pytest.raises(KeyError):
        result.covariance(params=[a, b, c], method="hesse")

    cov_mat_3 = result.covariance(params=[a, b, c])
    cov_mat_2 = result.covariance(params=[c, b])
    cov_dict = result.covariance(params=[a, b, c], as_dict=True)

    assert pytest.approx(hesse[a]["error"], rel=0.01) == np.sqrt(cov_dict[(a, a)])
    assert pytest.approx(hesse[a]["error"], rel=0.01) == np.sqrt(cov_mat_3[0, 0])

    assert pytest.approx(hesse[b]["error"], rel=0.01) == np.sqrt(cov_dict[(b, b)])
    assert pytest.approx(hesse[b]["error"], rel=0.01) == np.sqrt(cov_mat_3[1, 1])
    assert pytest.approx(hesse[b]["error"], rel=0.01) == np.sqrt(cov_mat_2[1, 1])

    assert pytest.approx(hesse[c]["error"], rel=0.01) == np.sqrt(cov_dict[(c, c)])
    assert pytest.approx(hesse[c]["error"], rel=0.01) == np.sqrt(cov_mat_3[2, 2])
    assert pytest.approx(hesse[c]["error"], rel=0.01) == np.sqrt(cov_mat_2[0, 0])

    if use_weights:
        rtol, atol = 0.15, 0.015
    else:
        rtol, atol = 0.05, 0.005

    cov_mat_3_np = result.covariance(params=[a, b, c], method="hesse_np")
    np.testing.assert_allclose(cov_mat_3, cov_mat_3_np, rtol=rtol, atol=atol)


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers, ids=minimizer_ids)
def test_correlation(minimizer_class_and_kwargs):
    results = create_fitresult(minimizer_class_and_kwargs=minimizer_class_and_kwargs)
    result = results["result"]
    hesse = result.hesse()
    a = results["a_param"]
    b = results["b_param"]
    c = results["c_param"]

    cor_mat = result.correlation(params=[a, b, c])
    cov_mat = result.covariance(params=[a, b, c])
    cor_dict = result.correlation(params=[a, b], as_dict=True)

    np.testing.assert_allclose(np.diag(cor_mat), 1.0)

    a_error = hesse[a]["error"]
    b_error = hesse[b]["error"]
    assert pytest.approx(cor_mat[0, 1], rel=0.01) == cov_mat[0, 1] / (a_error * b_error)
    assert pytest.approx(cor_dict[(a, b)], rel=0.01) == cov_mat[0, 1] / (
        a_error * b_error
    )


minimizers_test_errors = [
    (zfit.minimize.Minuit, {}, True),
]


@pytest.mark.parametrize(
    "minimizer_class_and_kwargs", minimizers_test_errors, ids=minimizer_ids
)
@pytest.mark.parametrize(
    "cl",
    [None, 0.683, 0.8, 0.95, 0.9],
    ids=["cldefault", "cl0683", "cl080", "cl095", "cl090"],
)
# @pytest.mark.timeout(1200)  # if stuck finding new minima
@pytest.mark.parametrize("extended", [True, False], ids=["extended", "not_extended"])
@pytest.mark.parametrize(
    "weights",
    [None, np.random.normal(1, 0.01, true_ntot)],
    ids=["no_weights", "weights"],
)
@pytest.mark.parametrize(
    "constraints", [False, True], ids=["no_constraints", "constraints"]
)
def test_errors(minimizer_class_and_kwargs, cl, weights, extended, constraints):
    n_max_trials = 5  # how often to try to find a new minimum
    results = create_fitresult(
        minimizer_class_and_kwargs=minimizer_class_and_kwargs,
        weights=weights,
        extended=extended,
        constraints=constraints,
    )
    result = results["result"]
    a = results["a_param"]
    b = results["b_param"]
    c = results["c_param"]
    params = list(result.params.keys())
    relerr = 0.03 if weights is None else 0.05

    for n_trial in range(n_max_trials):
        with zfit.run.set_autograd_mode(False):
            z_errors, new_result = result.errors(
                method="zfit_error",
                cl=cl,
                name="zfit_error",
            )
        minos_errors, _ = result.errors(
            method="minuit_minos", cl=cl, name="minuit_minos"
        )
        if new_result is None:
            break
        else:
            result = new_result
    else:  # no break occured
        assert False, "Always a new minimum was found, cannot perform test."

    for param in params:
        z_error_param = z_errors[param]
        minos_errors_param = minos_errors[param]
        if weights is not None and extended:
            continue  # TODO: fix this, somehow the uncertainty is not good for this case
        for dir in ["lower", "upper"]:
            assert (
                pytest.approx(z_error_param[dir], rel=relerr) == minos_errors_param[dir]
            )

    with pytest.raises(KeyError):
        result.errors(method="error")


@pytest.mark.flaky(reruns=3)
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers, ids=minimizer_ids)
def test_new_minimum(minimizer_class_and_kwargs):
    loss, params = create_loss(10000)

    minimizer_class, minimizer_kwargs, test_error = minimizer_class_and_kwargs
    minimizer = minimizer_class(**minimizer_kwargs)
    if isinstance(minimizer, zfit.minimize.NLoptLBFGSV1):
        return  # TODO: fix this, nlopt lbfgs cannot find the minimum when starting so close...
    a_param, b_param, c_param = params

    if test_error:
        b_param.floating = False
        b_param.set_value(3.7)
        c_param.floating = False
        result = minimizer.minimize(loss=loss)
        b_param.floating = True
        c_param.floating = True

        params_dict = {p: p.numpy() for p in params}
        hacked_result = FitResult(
            params=params_dict,
            edm=result.edm,
            fmin=result.fmin,
            info=result.info,
            loss=loss,
            status=result.status,
            converged=result.converged,
            valid=True,
            message="hacked for unittest",
            niter=999,
            criterion=None,
            minimizer=minimizer.copy(),
        )

        method = lambda **kwgs: compute_errors(covariance_method="hesse_np", **kwgs)

        errors, new_result = hacked_result.errors(
            params=params, method=method, name="interval"
        )

        assert new_result is not None

        assert hacked_result.valid is False
        for p in params:
            assert errors[p] == "Invalid, a new minimum was found."

        assert new_result.valid is True
        errors, _ = new_result.errors()
        for param in params:
            assert errors[param]["lower"] < 0
            assert errors[param]["upper"] > 0



#!/usr/bin/env python
#  Copyright (c) 2022 zfit

import argparse
import os
import re

import yaml

here = os.path.dirname(os.path.realpath(__file__))

parser = argparse.ArgumentParser(
    description="Replace arguments with central stored ones",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument("files", nargs="*", help="Files to be processed.")

parser.add_argument("--dry", action="store_true", help="Dry run WITHOUT replacing.")

cfg = parser.parse_args()

with open(here + "/argdocs.yaml") as replfile:
    replacements = yaml.load(replfile, Loader=yaml.Loader)

# Replace the target string
# auto_end_old = r'|@docend|'
for filepath in cfg.files:
    if not filepath.endswith(".py"):
        continue
    with open(filepath) as file:
        filedata = file.read()

    infile = False
    needs_replacement = False
    for param, replacement in replacements.items():
        replacement = replacement.rstrip("\n")
        while replacement[:1] == " ":  # we want to remove the whitespace
            replacement = replacement[1:]
        auto_start = rf"|@doc:{param}|"
        auto_end = rf"|@docend:{param}|"
        matches = re.findall(
            auto_start.replace("|", r"\|") + r".*?" + auto_end.replace("|", r"\|"),
            filedata,
            re.DOTALL,
        )

        if not matches:
            continue
        infile = True

        replacement_mod = f"{auto_start} {replacement} {auto_end}"

        for match in matches:
            if auto_start in match[len(auto_start) :]:  # sanity check
                raise ValueError(
                    f"Docstring formatting error,"
                    f" has more than one start until an end command: {match}"
                )
            if match != replacement_mod:
                needs_replacement = True
                filedata = filedata.replace(match, replacement_mod)

    # Write the file out again
    replace_msg = "replaced docs" if needs_replacement else "docs already there"
    filename = filepath.split("/")[-1]
    if infile:
        if cfg.dry:
            print(
                f"Match in {filename}, {replace_msg}, not writing to {filepath}, dry run."
            )
        else:
            if needs_replacement:
                with open(filepath, "w") as file:
                    file.write(filedata)
                print(f"Modified {filename}.")



#  Copyright (c) 2023 zfit
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Code was copied directly from tf_quant_finance
# The repository is deprecated and will be removed in the future.
"""Data class objects that works with tf.function."""

from typing import Type, Any
import attr

__all__ = ["dataclass"]


def dataclass(cls: Type[Any]) -> Type[Any]:
    """Creates a data class object compatible with `tf.function`.

    Modifies dunder methods of an input class with typed attributes to work as an
    input/output to `tf.function`, as well as a loop variable of
    `tf.while_loop`.

    An intended use case for this decorator is on top of a simple class definition
    with type annotated arguments like in the example below. It is not guaranteed
    that this decorator works with an arbitrary class.


    #### Examples

    ```python
    import tensorflow as tf
    import tf_quant_finance as tff

    @tff.utils.dataclass
    class Coords:
      x: tf.Tensor
      y: tf.Tensor

    @tf.function
    def fn(start_coords: Coords) -> Coords:
      def cond(it, _):
        return it < 10
      def body(it, coords):
        return it + 1, Coords(x=coords.x + 1, y=coords.y + 2)
      return tf.while_loop(cond, body, loop_vars=(0, start_coords))[1]

    start_coords = Coords(x=tf.constant(0), y=tf.constant(0))
    fn(start_coords)
    # Expected Coords(a=10, b=20)
    ```

    Args:
      cls: Input class object with type annotated arguments. The class should not
        have an init method defined. Class fields are treated as ordered in the
        same order as they appear in the class definition.


    Returns:
      Modified class that can be used as a `tf.function` input/output as well
      as a loop variable of `tf.function`. All typed arguments of the original
      class are treated as ordered in the same order as they appear in the class
      definition. All untyped arguments are ignored. Modified class modifies
      `len` and `iter` methods defined for the  class instances such that `len`
      returns the number of arguments, and `iter`  creates an iterator for the
      ordered argument values.
    """
    # Wrap the class with attr.s to ensure that the class can be an input/output
    # to a `tf.function`
    cls = attr.s(cls, auto_attribs=True)

    # Define __iter__ and __len__ method to ensure tf.while_loop compatibility
    def __iter__(self):  # pylint: disable=invalid-name
        # Note that self.__attrs_attrs__ is a tuple so the iteration order is fixed
        for item in self.__attrs_attrs__:
            name = item.name
            yield getattr(self, name)

    def __len__(self):  # pylint: disable=invalid-name
        return len(self.__attrs_attrs__)

    cls.__len__ = __len__
    cls.__iter__ = __iter__
    return cls
